/*
 * Programming in Java, Jacobs University, 2015
 *
 * This file implements a simple 'constrained property' of a bean, a
 * reusable software component that follows certain coding
 * conventions.  A constrained property is a special kind of a bound
 * property. It allows other pieces of code to listen for property
 * changes and in addition, a vetoable property change listener may
 * choose to veto a property change by throwing a PropertyVetoException.
 */ 

import java.beans.PropertyChangeSupport;
import java.beans.PropertyChangeListener;
import java.beans.VetoableChangeSupport;
import java.beans.VetoableChangeListener;
import java.beans.PropertyVetoException;

public class VetoableBean
    implements java.io.Serializable
{
    /*
     * A version number that needs to be bumped if there are
     * incompabible changes in the bean implementation.
     */

    private static final long serialVersionUID = 1L;

    private PropertyChangeSupport pcs;
    private VetoableChangeSupport vcs;
    public static final String PRICE = "price";
    private int price;

    public VetoableBean()
    {
        pcs = new PropertyChangeSupport(this);
        vcs = new VetoableChangeSupport(this);
    }

    public VetoableBean(int price)
    {
	this();
	this.price = price;
    }

    public void addPropertyChangeListener(PropertyChangeListener pcl)
    {
        pcs.addPropertyChangeListener(pcl);
    }

    public void removePropertyChangeListener(PropertyChangeListener pcl)
    {
        pcs.removePropertyChangeListener(pcl);
    }

    public void addVetoableChangeListener(VetoableChangeListener vcl)
    {
        vcs.addVetoableChangeListener(vcl);
    }

    public void removeVetoableChangeListener(VetoableChangeListener vcl)
    {
        vcs.removeVetoableChangeListener(vcl);
    }

    public void setPrice(int newValue)
	throws PropertyVetoException
    {
        if (newValue != price) {
            int oldValue = price;
            vcs.fireVetoableChange(PRICE, 
                                   new Integer(oldValue),
                                   new Integer(newValue));
            price = newValue;
            pcs.firePropertyChange(PRICE, 
                                   new Integer(oldValue),
                                   new Integer(newValue));
        }
    }

    public int getPrice()
    {
        return price;
    }

    public static void main(String[] args)
    {
	VetoableBean b = new VetoableBean(42);

	b.addPropertyChangeListener(e -> {
		if (e.getPropertyName() == BoundBean.PRICE) {
		    System.out.println("price change " + e.getOldValue()
				       + " -> " + e.getNewValue());
		}
	    });

	b.addVetoableChangeListener(e -> {
		if (e.getPropertyName() == VetoableBean.PRICE) {
		    int newValue = ((Integer) e.getNewValue()).intValue();
		    if (newValue < 0) {
			throw new PropertyVetoException("new price " + newValue+ " is negative", e);
		    }
		}
	    });
	
	b.addVetoableChangeListener(e -> {
		if (e.getPropertyName() == VetoableBean.PRICE) {
		    int oldValue = ((Integer) e.getOldValue()).intValue();
		    int newValue = ((Integer) e.getNewValue()).intValue();
		    int d = oldValue * 5 / 100;
		    if (newValue < oldValue - d || newValue > oldValue + d) {
			throw new PropertyVetoException("price change " + oldValue + " -> " + newValue + " > 5%", e);
		    }
		}
	    });

	int[] delta = { 2, -2, -44, 2, 2, 3, 4 };

	for (int d: delta) {
	    try {
		b.setPrice(b.getPrice() + d);
	    } catch (PropertyVetoException e) {
		System.err.println("veto: " + e.getMessage());
	    }
	}
    }
}
